/**
 *
 * Software to control Sky box via serial 'sky controller'
 *
 * Copyright (C) Joseph Heenan 2005-2008, all rights reserved.
 *
 * $Id: win-serial.c,v 1.5 2010-06-19 18:23:45 joseph Exp $
 *
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include <time.h>

#include "serial.h"

static void logprintf(const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);

    vprintf(fmt, ap);

    va_end(ap);
}

static void logError(DWORD lastError, serial_logfn logfn)
{
    LPVOID lpMsgBuf;

    FormatMessage(
                  FORMAT_MESSAGE_ALLOCATE_BUFFER |
                  FORMAT_MESSAGE_FROM_SYSTEM |
                  FORMAT_MESSAGE_IGNORE_INSERTS,
                  NULL,
                  lastError,
                  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                  (LPTSTR) &lpMsgBuf,
                  0, NULL );

	logfn("Windows error string: %s\n", lpMsgBuf);

    LocalFree(lpMsgBuf);
}

/**
 * @param port COM port - eg "COM4:"
 *
 * @param stopbits Not supported on windows - always 1 stop bit
 */
int serial_openportoverlapped(const char *port, int baudrate, int stopbits, BOOL overlapped, serial_logfn logfn)
{
    HANDLE       hPort;
    DCB          PortDCB;
    COMMTIMEOUTS CommTimeouts;
    char         portFname[128];

	if (!logfn)
		logfn = logprintf;

    _snprintf_s(portFname, sizeof(portFname), _TRUNCATE, "\\\\.\\%s", port);

    // Open the serial port.
    hPort = CreateFile (portFname, // Pointer to the name of the port
                        GENERIC_READ | GENERIC_WRITE,
                                      // Access (read-write) mode
                        0,            // Share mode
                        NULL,         // Pointer to the security attribute
                        OPEN_EXISTING,// How to open the serial port
                        overlapped?FILE_FLAG_OVERLAPPED:0, // flags & Port attributes
                        NULL);        // Handle to port with attribute
                                      // to copy

    if (hPort == INVALID_HANDLE_VALUE)
    {
        int dwError = GetLastError ();
        logfn("Unable to open serial port '%s' (%d)\n",
                port, dwError);
		logError(dwError, logfn);
        return -1;
    }

    memset(&PortDCB, 0, sizeof(PortDCB));
    // Initialize the DCBlength member.
    PortDCB.DCBlength = sizeof (DCB);

    // Get the default port setting information.
    GetCommState (hPort, &PortDCB);

    // Change the DCB structure settings.
    PortDCB.BaudRate = baudrate;          // Current baud
    PortDCB.fBinary = TRUE;               // Binary mode; no EOF check
    PortDCB.fParity = FALSE;              // No parity checking
    PortDCB.fOutxCtsFlow = FALSE;         // No CTS output flow control
    PortDCB.fOutxDsrFlow = FALSE;         // No DSR output flow control
    PortDCB.fDtrControl = DTR_CONTROL_ENABLE;
                                          // DTR flow control type
    PortDCB.fDsrSensitivity = FALSE;      // DSR sensitivity
    PortDCB.fTXContinueOnXoff = TRUE;     // XOFF continues Tx
    PortDCB.fOutX = FALSE;                // No XON/XOFF out flow control
    PortDCB.fInX = FALSE;                 // No XON/XOFF in flow control
    PortDCB.fErrorChar = FALSE;           // Disable error replacement
    PortDCB.fNull = FALSE;                // Disable null stripping
    PortDCB.fRtsControl = RTS_CONTROL_ENABLE;
                                          // RTS flow control
    PortDCB.fAbortOnError = FALSE;        // Do not abort reads/writes on
                                          // error
    PortDCB.ByteSize = 8;                 // Number of bits/byte, 4-8
    PortDCB.Parity = NOPARITY;            // 0-4=no,odd,even,mark,space
    PortDCB.StopBits = ONESTOPBIT;        // 0,1,2 = 1, 1.5, 2

    // Configure the port according to the specifications of the DCB
    // structure.
    if (!SetCommState (hPort, &PortDCB))
    {
        int dwError = GetLastError ();
        // Could not configure the serial port.
        logfn("Unable to configure the serial port (%d)\n",
                dwError);
		logError(dwError, logfn);
        return -1;
    }

    // Retrieve the time-out parameters for all read and write operations
    // on the port.
    memset(&CommTimeouts, 0, sizeof(CommTimeouts));
    GetCommTimeouts (hPort, &CommTimeouts);

    // Change the COMMTIMEOUTS structure settings.
    CommTimeouts.ReadIntervalTimeout = overlapped ? 100 : MAXDWORD; /* 100 ms between bytes */
    CommTimeouts.ReadTotalTimeoutMultiplier = 0;
    CommTimeouts.ReadTotalTimeoutConstant = overlapped ? 120 * 1000 : MAXDWORD; /* 120 seconds timeout for no data */
    CommTimeouts.WriteTotalTimeoutMultiplier = 10;
    CommTimeouts.WriteTotalTimeoutConstant = 1000;

    // Set the time-out parameters for all read and write operations
    // on the port.
    if (!SetCommTimeouts (hPort, &CommTimeouts))
    {
        int dwError = GetLastError ();
        // Could not set the time-out parameters.
        logfn("Unable to set the time-out parameters (%d)\n",
                dwError);
		logError(dwError, logfn);
        return -1;
    }

    return (int)hPort;
}

int serial_openport(const char *port, int baudrate, int stopbits, serial_logfn logfn)
{
	return serial_openportoverlapped(port, baudrate, stopbits, FALSE, logfn);
}

int serial_write(int hPort, const char *ptr, int bytes, serial_logfn logfn)
{
    DWORD send = -1;
    int ret;

	if (!logfn)
		logfn = logprintf;

    ret = WriteFile((HANDLE)hPort,        // Port handle
                    ptr,                  // Pointer to data to read
                    bytes,                // Number of bytes to read
                    &send,                // Pointer to number of bytes
                                          // written
                    NULL);                // Not overlapped
    if (ret == 0)
    {
        int dwError = GetLastError ();
        // Could not set the time-out parameters.
        logfn("serial write failed (%d)\n",
                dwError);
		logError(dwError, logfn);
        return -1;
    }

    return send;
}

int serial_read(int hPort, unsigned char *ptr, int bytes, serial_logfn logfn)
{
    DWORD cnt = -1;
    int ret;
    HANDLE h = (HANDLE)hPort;

	if (!logfn)
		logfn = logprintf;

    ret = ReadFile(h,                   // Port handle
                   ptr,                  // Pointer to data to read
                   bytes,                // Number of bytes to read
                   &cnt,                 // Pointer to number of bytes
                                         // written
                   NULL);                // Not overlapped
    if (ret == 0)
    {
        logfn("Failed to read from serial port\n");
        exit(1);
    }

    return cnt;
}

int serial_readtimeout(int fd, unsigned char *ptr, int bytes,
                       int timeout, serial_logfn logfn)
{
    DWORD cnt = -1;
    int ret;
    HANDLE hPort = (HANDLE) fd;
    COMMTIMEOUTS CommTimeouts;
    COMMTIMEOUTS OrigCommTimeouts;

	if (!logfn)
		logfn = logprintf;

    // Retrieve the time-out parameters for all read and write operations
    // on the port.
    memset(&CommTimeouts, 0, sizeof(CommTimeouts));
    GetCommTimeouts (hPort, &CommTimeouts);
    OrigCommTimeouts = CommTimeouts;

    // Change the COMMTIMEOUTS structure settings.
    CommTimeouts.ReadIntervalTimeout = MAXDWORD;
    CommTimeouts.ReadTotalTimeoutMultiplier = 0;
    CommTimeouts.ReadTotalTimeoutConstant = timeout; /* 100 seconds timeout */
    CommTimeouts.WriteTotalTimeoutMultiplier = 10;
    CommTimeouts.WriteTotalTimeoutConstant = 1000;

    // Set the time-out parameters for all read and write operations
    // on the port.
    if (!SetCommTimeouts (hPort, &CommTimeouts))
    {
        int dwError = GetLastError ();
        // Could not set the time-out parameters.
        logfn("Unable to set the time-out parameters (%d)\n",
              dwError);
        return -1;
    }

    ret = ReadFile(hPort,                // Port handle
                   ptr,                  // Pointer to data to read
                   bytes,                // Number of bytes to read
                   &cnt,                 // Pointer to number of bytes
                                         // written
                   NULL);                // Not overlapped
    if (ret == 0)
    {
        int dwError = GetLastError ();
        logfn("Failed to read from serial port (%d)\n",
                dwError);
        exit(1);
    }
    if (cnt == 0)
    {
        /* timeout must have been hit */
        return -2;
    }

    /* restore previous timeouts */
    if (!SetCommTimeouts (hPort, &OrigCommTimeouts))
    {
        int dwError = GetLastError ();
        // Could not set the time-out parameters.
        logfn("Unable to set the time-out parameters (%d)\n",
                dwError);
        return -1;
    }

    return cnt;
}


void serial_close(int fd, serial_logfn logfn)
{
    BOOL ret;
    HANDLE h = (HANDLE)fd;

	if (!logfn)
		logfn = logprintf;

	ret = CloseHandle(h);
    if (ret == 0)
    {
        int dwError = GetLastError ();
        logfn("CloseHandle failed (%d)\n",
                dwError);

    }
}
